from abc import abstractmethod, ABC

import sqlite3
from google.cloud.bigquery import Client


class DatabaseClient(ABC):
    def __init__(self, client):
        self.client = client
        super(DatabaseClient, self).__init__()

    @abstractmethod
    def execute(self, query: str):
        pass


class BigQueryDatabaseClient(DatabaseClient):
    """ If running locally, run the following commandline to authenticate yourself:
    gcloud auth application-default login
    """

    def __init__(self, client=None, credentials_file=None):
        if client is not None:
            super(BigQueryDatabaseClient, self).__init__(client)
        else:
            if credentials_file is not None:
                bigquery_client = Client.from_service_account_json(credentials_file)
            else:
                raise ValueError("BigQueryDatabaseClient requires a client or a credentials_file.")
            super(BigQueryDatabaseClient, self).__init__(bigquery_client)

    def execute(self, query: str):
        query_job = self.client.query(query)  # API request
        rows = query_job.result()  # Waits for query to finish
        return rows


class SqLiteDatabaseClient(DatabaseClient):
    def __init__(self, client=None, db_file=None):
        if client is not None:
            super(SqLiteDatabaseClient, self).__init__(client)
        else:
            if db_file is not None:
                super(SqLiteDatabaseClient, self).__init__(sqlite3.connect(db_file).cursor())
            else:
                raise ValueError("SqLiteDatabaseClient requires a client or a db_file.")

    def execute(self, query: str):
        return self.client.execute(query)


if '__main__' == __name__:
    credentials_file = '/Users/kyuksel/ml4h/bigquery-viewer-credentials.json'
    db_client = BigQueryDatabaseClient(credentials_file=credentials_file)

    dataset = 'broad-ml4cvd.ukbb7089_r10data'

    dictionary_table = f"`{dataset}.dictionary`"
    phenotype_table = f"`{dataset}.phenotype`"
    coding_table = f"`{dataset}.coding`"

    fid = 20001
    fids = [3143, 3144]
    sample_id = 2907043

    job_title_field_id = 22600
    icd10_field = 41202
    query = \
        f"SELECT value FROM {phenotype_table} WHERE fieldid={icd10_field} AND sample_id={sample_id}"

    rows = db_client.execute(query.format())
    for row in rows:
        print(row)



"""Methods for integration of plots of mri data processed to 3D tensors from within notebooks."""

from collections import OrderedDict
from enum import Enum, auto
import os
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Union

from IPython.display import display
from IPython.display import HTML
import numpy as np
import h5py
import ipywidgets as widgets
import matplotlib.pyplot as plt
from ml4h.runtime_data_defines import get_mri_hd5_folder
import ml4h.tensormap.ukb.mri as ukb_mri
import ml4h.tensormap.ukb.mri_vtk as ukb_mri_vtk
from ml4h.TensorMap import Interpretation, TensorMap
import tensorflow as tf

# Discover applicable TensorMaps.
MRI_TMAPS = {
    key: value for key, value in ukb_mri.__dict__.items() if isinstance(value, TensorMap)
    and value.interpretation == Interpretation.CONTINUOUS and value.axes() == 3
}
MRI_TMAPS.update(
    {
        key: value for key, value in ukb_mri_vtk.__dict__.items()
        if isinstance(value, TensorMap) and value.interpretation == Interpretation.CONTINUOUS and value.axes() == 3
    },
)

MIN_IMAGE_WIDTH = 8
DEFAULT_IMAGE_WIDTH = 12
MAX_IMAGE_WIDTH = 24

MIN_COLOR_RANGE = 0
MAX_COLOR_RANGE = 6000


class PlotType(Enum):
  INTERACTIVE = auto()
  PANEL = auto()


class TensorMapCache:
  """Cache the tensor to display for reuse when re-plotting the same TMAP with different plot parameters."""

  def __init__(self, hd5: Dict[str, Any], tmap: TensorMap):
    self.hd5 = hd5
    self.tmap: Optional[TensorMap] = None
    self.tensor = None
    _ = self.get(tmap)

  def get(self, tmap: TensorMap) -> np.array:
    if self.tmap != tmap:
      self.tensor = tmap.tensor_from_file(tmap, self.hd5)
      self.tmap = tmap
    return self.tensor


def choose_mri_tmap(
    sample_id: Union[int, str], folder: Optional[str] = None, tmap: Optional[TensorMap] = None,
    default_tmaps: Dict[str, TensorMap] = MRI_TMAPS,
) -> None:
  """Render widgets and plots for MRI tensors.

  Args:
    sample_id: The id of the sample to retrieve.
    folder: The local or Cloud Storage folder under which the files reside.
    tmap: The TensorMap for the 3D MRI tensor to visualize.
    default_tmaps: Other TensorMaps to offer for visualization, if present in the hd5.
  """
  if folder is None:
    folder = get_mri_hd5_folder(sample_id)

  with tempfile.TemporaryDirectory() as tmpdirname:
    sample_hd5 = str(sample_id) + '.hd5'
    local_path = os.path.join(tmpdirname, sample_hd5)
    try:
      tf.io.gfile.copy(src=os.path.join(folder, sample_hd5), dst=local_path)
      hd5 = h5py.File(local_path, mode='r')
    except (tf.errors.NotFoundError, tf.errors.PermissionDeniedError) as e:
      display(
          HTML(f'''<div class="alert alert-block alert-danger">
      <b>Warning:</b> MRI HD5 file not available for sample {sample_id} in folder {folder}:
      <hr><p><pre>{e.message}</pre></p>
      Use the <kbd>folder</kbd> parameter to read HD5s from a different local directory or Cloud Storage bucket.
      </div>'''),
      )
      return

    sample_tmaps = OrderedDict()
    # Add the passed tmap parameter, if it is present in this hd5.
    if tmap:
      if tmap.hd5_key_guess() in hd5:
        if len(tmap.shape) == 3:
          sample_tmaps[tmap.name] = tmap
        else:
          print(f'{tmap} is not a 3D tensor, skipping it')
      else:
        print(f'{tmap} is not available in {sample_id}')
    # Also discover applicable TensorMaps for this particular sample's HD5 file.
    sample_tmaps.update({n: t for n, t in sorted(default_tmaps.items(), key=lambda t: t[0]) if t.hd5_key_guess() in hd5})

    if not sample_tmaps:
      display(
          HTML(f'''<div class="alert alert-block alert-danger">
      Neither {tmap.name} nor any of {default_tmaps.keys()} are present in this HD5 for sample {sample_id} in {folder}.
      Use the tmap parameter to try a different TensorMap or the folder parameter to try a different hd5 for the sample.
      </div>'''),
      )
      return

    default_tmap_value = next(iter(sample_tmaps.values()))
    # Display the middle instance by default in the interactive view.
    default_instance_value, max_instance_value = compute_instance_range(default_tmap_value)
    default_vmin_value, default_vmax_value = compute_color_range(hd5, default_tmap_value)

    tmap_chooser = widgets.Dropdown(
        options=list(sample_tmaps.items()),
        value=default_tmap_value,
        description='Choose the MRI tensor TMAP name to visualize:',
        style={'description_width': 'initial'},
        layout=widgets.Layout(width='900px'),
    )
    fig_width_chooser = widgets.IntSlider(
        continuous_update=False,
        value=DEFAULT_IMAGE_WIDTH,
        min=MIN_IMAGE_WIDTH,
        max=MAX_IMAGE_WIDTH,
        description='Desired width of figure (height computed using input data)',
        style={'description_width': 'initial'},
        layout=widgets.Layout(width='600px'),
    )
    transpose_chooser = widgets.Checkbox(
        description='Whether to transpose the images.',
        style={'description_width': 'initial'},
        layout=widgets.Layout(width='300px'),
    )
    flip_chooser = widgets.Checkbox(
        description='Whether to flip the images.',
        style={'description_width': 'initial'},
        layout=transpose_chooser.layout,
    )
    plot_type_chooser = widgets.RadioButtons(
        options=[('interactive animation', PlotType.INTERACTIVE), ('panel grid', PlotType.PANEL)],
        description='Plot type',
        style={'description_width': 'initial'},
        layout=transpose_chooser.layout,
    )
    instance_chooser = widgets.IntSlider(
        continuous_update=True,
        value=default_instance_value,
        min=1,
        max=max_instance_value,
        description='Instance to display (click on slider, then use left/right arrows):',
        style={'description_width': 'initial'},
        layout=widgets.Layout(width='900px'),
    )
    color_range_chooser = widgets.IntRangeSlider(
        continuous_update=False,
        value=[default_vmin_value, default_vmax_value],
        min=MIN_COLOR_RANGE,
        max=MAX_COLOR_RANGE,
        description='Color range:',
        style={'description_width': 'initial'},
        layout=fig_width_chooser.layout,
    )
    viz_controls_ui = widgets.VBox(
        [
            widgets.HTML('<h3>Visualization controls</h3>'),
            tmap_chooser,
            widgets.HBox([transpose_chooser, fig_width_chooser]),
            widgets.HBox([flip_chooser, color_range_chooser]),
            widgets.HBox([plot_type_chooser, instance_chooser]),
        ],
        layout=widgets.Layout(width='auto', border='solid 1px grey'),
    )
    tmap_cache = TensorMapCache(hd5=hd5, tmap=tmap_chooser.value)
    viz_controls_output = widgets.interactive_output(
        plot_mri_tmap,
        {
            'sample_id': widgets.fixed(sample_id),
            'tmap_cache': widgets.fixed(tmap_cache),
            'tmap': tmap_chooser,
            'plot_type': plot_type_chooser,
            'instance': instance_chooser,
            'color_range': color_range_chooser,
            'transpose': transpose_chooser,
            'flip': flip_chooser,
            'fig_width': fig_width_chooser,
        },
    )

    def on_tmap_value_change(change):
      """When the TMAP name changes, update the widgets to the proper ranges for the new TMAP name."""
      color_range_chooser.value = compute_color_range(hd5, change['new'])
      instance_chooser.value, instance_chooser.max = compute_instance_range(change['new'])

    def on_plot_type_change(change):
      """Only show controls applicable to the plot type."""
      if change['new'] == PlotType.INTERACTIVE:
        instance_chooser.layout.visibility = 'visible'
      else:
        instance_chooser.layout.visibility = 'hidden'

    tmap_chooser.observe(on_tmap_value_change, names='value')
    plot_type_chooser.observe(on_plot_type_change, names='value')
    display(viz_controls_ui, viz_controls_output)


def compute_color_range(hd5: Dict[str, Any], tmap: TensorMap) -> List[int]:
  """Compute the mean values for the color ranges of instances in the MRI series."""
  mri_tensor = tmap.tensor_from_file(tmap, hd5)
  vmin = np.mean([np.min(mri_tensor[:, :, i]) for i in range(0, mri_tensor.shape[2])])
  vmax = np.mean([np.max(mri_tensor[:, :, i]) for i in range(0, mri_tensor.shape[2])])
  return [vmin, vmax]


def compute_instance_range(tmap: TensorMap) -> Tuple[int, int]:
  """Compute middle and max instances."""
  middle_instance = int(tmap.shape[2] / 2)
  max_instance = tmap.shape[2]
  return (middle_instance, max_instance)


def plot_mri_tmap(
    sample_id: Union[int, str], tmap_cache: TensorMapCache, tmap: TensorMap, plot_type: PlotType,
    instance: int, color_range: Tuple[int, int], transpose: bool, flip: bool, fig_width: int,
) -> None:
  """Visualize the applicable MRI series within this HD5 file.

  Args:
    sample_id: The local or Cloud Storage path to the MRI file.
    tmap_cache: The cache from which to retrieve the tensor to be plotted.
    tmap: The chosen TensorMap for the MRI series.
    plot_type: Whether to display instances interactively or in a panel view.
    instance: The particular instance to display, if interactive.
    color_range: Array of minimum and maximum value for the color range.
    transpose: Whether to transpose the images.
    flip: Whether to flip the image on its vertical axis
    fig_width: The desired width of the figure. Note that height computed as
      the proportion of the width based on the data to be plotted.
  """
  title_prefix = f'{tmap.name} from MRI {sample_id}'
  mri_tensor = tmap_cache.get(tmap)
  if plot_type == PlotType.INTERACTIVE:
    plot_mri_tensor_as_animation(
        mri_tensor=mri_tensor,
        instance=instance,
        vmin=color_range[0],
        vmax=color_range[1],
        transpose=transpose,
        flip=flip,
        fig_width=fig_width,
        title_prefix=title_prefix,
    )
  elif plot_type == PlotType.PANEL:
    # Note: this print statement causes the current image to go away while the new one is rendering.
    # Do this intensionally for the panel plots, because they are a bit slower to render. It serves
    # a purpose similar to a progress bar.
    print(f'Rendering {title_prefix} . . .')
    plot_mri_tensor_as_panels(
        mri_tensor=mri_tensor,
        vmin=color_range[0],
        vmax=color_range[1],
        transpose=transpose,
        flip=flip,
        fig_width=fig_width,
        title_prefix=title_prefix,
    )
  else:
    HTML(f'''<div class="alert alert-block alert-danger">Invalid plot type: {plot_type}</div>''')


def plot_mri_tensor_as_panels(
    mri_tensor: np.array, vmin: int, vmax: int, transpose: bool = False, flip: bool = False,
    fig_width: int = DEFAULT_IMAGE_WIDTH, title_prefix: str = '',
) -> None:
  """Visualize an MRI series from a 3D tensor as a panel of static plots.

  Args:
    mri_tensor: The MRI 3D tensor.
    vmin: Minimum value for the color range.
    vmax: Maximum value for the color range.
    transpose: Whether to transpose the image.
    flip: Whether to flip the image on its vertical axis.
    fig_width: The desired width of the figure, note that height computed as
      the proportion of the width based on the data to be plotted.
    title_prefix: Text to display as the initial portion of the plot title.
  """
  cols = 5
  rows = int(np.ceil(mri_tensor.shape[2] / 5.0))
  if transpose:
    height = mri_tensor.shape[1]
    width = mri_tensor.shape[0]
  else:
    height = mri_tensor.shape[0]
    width = mri_tensor.shape[1]

  fig_height = int(np.ceil(fig_width * ((rows * height)/(cols * width))))
  fig, axes = plt.subplots(rows, cols, figsize=(fig_width, fig_height), facecolor='beige')
  for i in range(0, mri_tensor.shape[2]):
    col = i % cols
    row = i // cols
    pixels = mri_tensor[:, :, i]
    if transpose:
      pixels = pixels.T
    if flip:
      pixels = np.flip(pixels, 1)
    axes[row, col].imshow(pixels, cmap='gray', vmin=vmin, vmax=vmax)
    axes[row, col].set_yticklabels([])
    axes[row, col].set_xticklabels([])
  fig.suptitle(
      f'{title_prefix}\nColor range: {vmin}-{vmax}, Transpose: {transpose}, Flip: {flip}, Figure size:{fig_width}x{fig_height}',  # pylint: disable=line-too-long
      fontsize=fig_width,
  )
  fig.subplots_adjust(
      top=0.96,    # the top of the subplots of the figure
      wspace=0.1,  # the amount of width reserved for space between subplots,
                   # expressed as a fraction of the average axis width
      hspace=0.1,  # the amount of height reserved for space between subplots,
                   # expressed as a fraction of the average axis height
  )


def plot_mri_tensor_as_animation(
    mri_tensor: np.array, instance: int, vmin: int, vmax: int,
    transpose: bool = False, flip: bool = False,
    fig_width: int = DEFAULT_IMAGE_WIDTH, title_prefix: str = '',
) -> None:
  """Visualize an MRI series from a 3D tensor as an animation rendered one panel at a time.

  Args:
    mri_tensor: The MRI 3D tensor.
    instance: The particular instance to display.
    vmin: Minimum value for the color range.
    vmax: Maximum value for the color range.
    transpose: Whether to transpose the image.
    flip: Whether to flip the image on its vertical axis.
    fig_width: The desired width of the figure, note that height computed as
      the proportion of the width based on the data to be plotted.
    title_prefix: Text to display as the initial portion of the plot title.
  """
  if mri_tensor.shape[2] < instance:
    pixels = mri_tensor[:, :, -1]
    print(f'Instance {str(instance)} not available for {title_prefix}, using final instance instead.')
  else:
    pixels = mri_tensor[:, :, instance - 1]

  if transpose:
    pixels = pixels.T
  if flip:
    pixels = np.flip(pixels, 1)

  height = pixels.shape[0]
  width = pixels.shape[1]
  fig_height = int(np.ceil(fig_width * (height/width)))

  _, ax = plt.subplots(figsize=(fig_width, fig_height), facecolor='beige')
  ax.imshow(pixels, cmap='gray', vmin=vmin, vmax=vmax)
  ax.set_title(
      f'{title_prefix}, Instance: {instance}\nColor range: {vmin}-{vmax}, Transpose: {transpose}, Flip: {flip}, Figure size:{fig_width}x{fig_height}',  # pylint: disable=line-too-long
      fontsize=fig_width,
  )
  ax.set_yticklabels([])
  ax.set_xticklabels([])



from tensorflow.keras.losses import mean_squared_error

from ml4h.TensorMap import TensorMap, Interpretation
from ml4h.defines import StorageType
from ml4h.metrics import weighted_crossentropy, abs_pearson, pearson

diploid_cm = {'homozygous_reference': 0, 'heterozygous': 1, 'homozygous_variant': 2}
rs3829740 = TensorMap('rs3829740', Interpretation.CATEGORICAL, channel_map=diploid_cm)
rs2234962 = TensorMap('rs2234962', Interpretation.CATEGORICAL, channel_map=diploid_cm)
rs2042995 = TensorMap('rs2042995', Interpretation.CATEGORICAL, channel_map=diploid_cm)

rs3829740_weighted = TensorMap(
    'rs3829740', Interpretation.CATEGORICAL, channel_map=diploid_cm,
    loss=weighted_crossentropy([1, 1, 1.5], 'rs3829740'),
)
rs2234962_weighted = TensorMap(
    'rs2234962', Interpretation.CATEGORICAL, channel_map=diploid_cm,
    loss=weighted_crossentropy([.8, 1, 1.5], 'rs2234962'),
)
rs2042995_weighted = TensorMap(
    'rs2042995', Interpretation.CATEGORICAL, channel_map=diploid_cm,
    loss=weighted_crossentropy([.6, 1.5, 2], 'rs2042995'),
)

# KCNJ5 = TensorMap('KCNJ5', Interpretation.CONTIG, channel_map={'position': 0, 'genotype': 1})
# KCNH2 = TensorMap('KCNH2', Interpretation.CONTIG, channel_map={'position': 0, 'genotype': 1})
# SCN5A = TensorMap('SCN5A', Interpretation.CONTIG, channel_map={'position': 0, 'genotype': 1})
# TTN = TensorMap('TTN', Interpretation.CONTIG, channel_map={'position': 0, 'genotype': 1})

akap9_lof = TensorMap('AKAP9', Interpretation.CATEGORICAL, channel_map={'no_akap9_lof': 0, 'akap9_lof': 1})
dsc2_lof = TensorMap('DSC2', Interpretation.CATEGORICAL, channel_map={'no_dsc2_lof': 0, 'dsc2_lof': 1})
ryr2_lof = TensorMap('RYR2', Interpretation.CATEGORICAL, channel_map={'no_ryr2_lof': 0, 'ryr2_lof': 1})
ttn_lof = TensorMap('ttn_lof', Interpretation.CATEGORICAL, channel_map={'no_ttn_lof': 0, 'ttn_lof': 1})

TTN_LOF = TensorMap(
    'TTN', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_INDEX, channel_map={'no_TTN_LOF': 0, 'TTN_LOF': 1},
)
LMNA_LOF = TensorMap(
    'LMNA', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_INDEX, channel_map={'no_LMNA_LOF': 0, 'LMNA_LOF': 1},
)
PKP2_LOF = TensorMap(
    'PKP2', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_INDEX, channel_map={'no_PKP2_LOF': 0, 'PKP2_LOF': 1},
)
SCN5A_LOF = TensorMap(
    'SCN5A', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_INDEX, channel_map={'no_SCN5A_LOF': 0, 'SCN5A_LOF': 1},
)
KCNQ1_LOF = TensorMap(
    'KCNQ1', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_INDEX, channel_map={'no_KCNQ1_LOF': 0, 'KCNQ1_LOF': 1},
)
KCNH2_LOF = TensorMap(
    'KCNH2', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_INDEX, channel_map={'no_KCNH2_LOF': 0, 'KCNH2_LOF': 1},
)
KCNJ5_LOF = TensorMap(
    'KCNJ5', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_INDEX, channel_map={'no_KCNJ5_LOF': 0, 'KCNJ5_LOF': 1},
)
MYBPC3_LOF = TensorMap(
    'MYBPC3', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_INDEX, channel_map={'no_MYBPC3_LOF': 0, 'MYBPC3_LOF': 1},
)

TTN_LOF_weighted = TensorMap(
    'TTN', Interpretation.CATEGORICAL, path_prefix='categorical', loss=weighted_crossentropy([0.2, 15.0]),
    storage_type=StorageType.CATEGORICAL_INDEX, channel_map={'no_TTN_LOF': 0, 'TTN_LOF': 1},
)

ttntv = TensorMap(
    'has_ttntv', Interpretation.CATEGORICAL, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_INDEX,
    channel_map={'no_ttntv': 0, 'has_ttntv': 1},
)

bsa_mosteller = TensorMap(
    'bsa_mosteller', Interpretation.CONTINUOUS,
    normalization={'mean': 1.8894831981880114, 'std': 0.22169301057810176}, loss='log_cosh',
    channel_map={'bsa_mosteller': 0},
)
bsa_dubois = TensorMap(
    'bsa_dubois', Interpretation.CONTINUOUS,
    normalization={'mean': 1.8671809970639703, 'std': 0.20913930961120797}, loss='log_cosh',
    channel_map={'bsa_dubois': 0},
)

ancestry_pca_40 = TensorMap(
    'ancestry_pca_40', Interpretation.CONTINUOUS, shape=(40,), loss='log_cosh', path_prefix='continuous',
    channel_map={
        '22009_Genetic-principal-components_0_1': 0, '22009_Genetic-principal-components_0_2': 1,
        '22009_Genetic-principal-components_0_3': 2, '22009_Genetic-principal-components_0_4': 3,
        '22009_Genetic-principal-components_0_5': 4, '22009_Genetic-principal-components_0_6': 5,
        '22009_Genetic-principal-components_0_7': 6, '22009_Genetic-principal-components_0_8': 7,
        '22009_Genetic-principal-components_0_9': 8, '22009_Genetic-principal-components_0_10': 9,
        '22009_Genetic-principal-components_0_11': 10, '22009_Genetic-principal-components_0_12': 11,
        '22009_Genetic-principal-components_0_13': 12, '22009_Genetic-principal-components_0_14': 13,
        '22009_Genetic-principal-components_0_15': 14, '22009_Genetic-principal-components_0_16': 15,
        '22009_Genetic-principal-components_0_17': 16, '22009_Genetic-principal-components_0_18': 17,
        '22009_Genetic-principal-components_0_19': 18, '22009_Genetic-principal-components_0_20': 19,
        '22009_Genetic-principal-components_0_21': 20, '22009_Genetic-principal-components_0_22': 21,
        '22009_Genetic-principal-components_0_23': 22, '22009_Genetic-principal-components_0_24': 23,
        '22009_Genetic-principal-components_0_25': 24, '22009_Genetic-principal-components_0_26': 25,
        '22009_Genetic-principal-components_0_27': 26, '22009_Genetic-principal-components_0_28': 27,
        '22009_Genetic-principal-components_0_29': 28, '22009_Genetic-principal-components_0_30': 29,
        '22009_Genetic-principal-components_0_31': 30, '22009_Genetic-principal-components_0_32': 31,
        '22009_Genetic-principal-components_0_33': 32, '22009_Genetic-principal-components_0_34': 33,
        '22009_Genetic-principal-components_0_35': 34, '22009_Genetic-principal-components_0_36': 35,
        '22009_Genetic-principal-components_0_37': 36, '22009_Genetic-principal-components_0_38': 37,
        '22009_Genetic-principal-components_0_39': 38, '22009_Genetic-principal-components_0_40': 39,
    },
)

ancestry_pca_43 = TensorMap(
    'ancestry_pca_43', Interpretation.CONTINUOUS, shape=(43,), loss='log_cosh', path_prefix='continuous',
    channel_map={
        '22005_Missingness_0_0': 0, '22004_Heterozygosity-PCA-corrected_0_0': 1, '22003_Heterozygosity_0_0': 2,
        '22009_Genetic-principal-components_0_1': 3, '22009_Genetic-principal-components_0_2': 4,
        '22009_Genetic-principal-components_0_3': 5, '22009_Genetic-principal-components_0_4': 6,
        '22009_Genetic-principal-components_0_5': 7, '22009_Genetic-principal-components_0_6': 8,
        '22009_Genetic-principal-components_0_7': 9, '22009_Genetic-principal-components_0_8': 10,
        '22009_Genetic-principal-components_0_9': 11, '22009_Genetic-principal-components_0_10': 12,
        '22009_Genetic-principal-components_0_11': 13, '22009_Genetic-principal-components_0_12': 14,
        '22009_Genetic-principal-components_0_13': 15, '22009_Genetic-principal-components_0_14': 16,
        '22009_Genetic-principal-components_0_15': 17, '22009_Genetic-principal-components_0_16': 18,
        '22009_Genetic-principal-components_0_17': 19, '22009_Genetic-principal-components_0_18': 20,
        '22009_Genetic-principal-components_0_19': 21, '22009_Genetic-principal-components_0_20': 22,
        '22009_Genetic-principal-components_0_21': 23, '22009_Genetic-principal-components_0_22': 24,
        '22009_Genetic-principal-components_0_23': 25, '22009_Genetic-principal-components_0_24': 26,
        '22009_Genetic-principal-components_0_25': 27, '22009_Genetic-principal-components_0_26': 28,
        '22009_Genetic-principal-components_0_27': 29, '22009_Genetic-principal-components_0_28': 30,
        '22009_Genetic-principal-components_0_29': 31, '22009_Genetic-principal-components_0_30': 32,
        '22009_Genetic-principal-components_0_31': 33, '22009_Genetic-principal-components_0_32': 34,
        '22009_Genetic-principal-components_0_33': 35, '22009_Genetic-principal-components_0_34': 36,
        '22009_Genetic-principal-components_0_35': 37, '22009_Genetic-principal-components_0_36': 38,
        '22009_Genetic-principal-components_0_37': 39, '22009_Genetic-principal-components_0_38': 40,
        '22009_Genetic-principal-components_0_39': 41, '22009_Genetic-principal-components_0_40': 42,
    },
)

genetic_pca_1 = TensorMap(
    '22009_Genetic-principal-components_0_1', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_1': 0},
)
genetic_pca_2 = TensorMap(
    '22009_Genetic-principal-components_0_2', Interpretation.CONTINUOUS, path_prefix='continuous',
    # normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_2': 0},
)
genetic_pca_3 = TensorMap(
    '22009_Genetic-principal-components_0_3', Interpretation.CONTINUOUS, path_prefix='continuous',
    # normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_3': 0},
)
genetic_pca_4 = TensorMap(
    '22009_Genetic-principal-components_0_4', Interpretation.CONTINUOUS, path_prefix='continuous',
    # normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_4': 0},
)
genetic_pca_5 = TensorMap(
    '22009_Genetic-principal-components_0_5', Interpretation.CONTINUOUS, path_prefix='continuous',
    # normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_5': 0},
)
genetic_pca_6 = TensorMap(
    '22009_Genetic-principal-components_0_6', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_6': 0},
)

genetic_pca_7 = TensorMap(
    '22009_Genetic-principal-components_0_7', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_7': 0},
)

genetic_pca_8 = TensorMap(
    '22009_Genetic-principal-components_0_8', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_8': 0},
)

genetic_pca_9 = TensorMap(
    '22009_Genetic-principal-components_0_9', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_9': 0},
)

genetic_pca_10 = TensorMap(
    '22009_Genetic-principal-components_0_10', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_10': 0},
)

genetic_pca_11 = TensorMap(
    '22009_Genetic-principal-components_0_11', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_11': 0},
)

genetic_pca_12 = TensorMap(
    '22009_Genetic-principal-components_0_12', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_12': 0},
)

genetic_pca_13 = TensorMap(
    '22009_Genetic-principal-components_0_13', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_13': 0},
)

genetic_pca_14 = TensorMap(
    '22009_Genetic-principal-components_0_14', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_14': 0},
)

genetic_pca_15 = TensorMap(
    '22009_Genetic-principal-components_0_15', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_15': 0},
)

genetic_pca_16 = TensorMap(
    '22009_Genetic-principal-components_0_16', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_16': 0},
)

genetic_pca_17 = TensorMap(
    '22009_Genetic-principal-components_0_17', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_17': 0},
)

genetic_pca_18 = TensorMap(
    '22009_Genetic-principal-components_0_18', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_18': 0},
)

genetic_pca_19 = TensorMap(
    '22009_Genetic-principal-components_0_19', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_19': 0},
)

genetic_pca_20 = TensorMap(
    '22009_Genetic-principal-components_0_20', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_20': 0},
)

genetic_pca_21 = TensorMap(
    '22009_Genetic-principal-components_0_21', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_21': 0},
)

genetic_pca_22 = TensorMap(
    '22009_Genetic-principal-components_0_22', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_22': 0},
)

genetic_pca_23 = TensorMap(
    '22009_Genetic-principal-components_0_23', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_23': 0},
)

genetic_pca_24 = TensorMap(
    '22009_Genetic-principal-components_0_24', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_24': 0},
)

genetic_pca_25 = TensorMap(
    '22009_Genetic-principal-components_0_25', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_25': 0},
)

genetic_pca_26 = TensorMap(
    '22009_Genetic-principal-components_0_26', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_26': 0},
)

genetic_pca_27 = TensorMap(
    '22009_Genetic-principal-components_0_27', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_27': 0},
)

genetic_pca_28 = TensorMap(
    '22009_Genetic-principal-components_0_28', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_28': 0},
)

genetic_pca_29 = TensorMap(
    '22009_Genetic-principal-components_0_29', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_29': 0},
)

genetic_pca_30 = TensorMap(
    '22009_Genetic-principal-components_0_30', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_30': 0},
)

genetic_pca_31 = TensorMap(
    '22009_Genetic-principal-components_0_31', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_31': 0},
)

genetic_pca_32 = TensorMap(
    '22009_Genetic-principal-components_0_32', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_32': 0},
)

genetic_pca_33 = TensorMap(
    '22009_Genetic-principal-components_0_33', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_33': 0},
)

genetic_pca_34 = TensorMap(
    '22009_Genetic-principal-components_0_34', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_34': 0},
)

genetic_pca_35 = TensorMap(
    '22009_Genetic-principal-components_0_35', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_35': 0},
)

genetic_pca_36 = TensorMap(
    '22009_Genetic-principal-components_0_36', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_36': 0},
)

genetic_pca_37 = TensorMap(
    '22009_Genetic-principal-components_0_37', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_37': 0},
)

genetic_pca_38 = TensorMap(
    '22009_Genetic-principal-components_0_38', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_38': 0},
)

genetic_pca_39 = TensorMap(
    '22009_Genetic-principal-components_0_39', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_39': 0},
)

genetic_pca_40 = TensorMap(
    '22009_Genetic-principal-components_0_40', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss='log_cosh', activation='linear', channel_map={'22009_Genetic-principal-components_0_40': 0},
)

offset = 192
genetic_pca_1_partition = TensorMap(
    '22009_Genetic-principal-components_0_1', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_1': 0},
)
genetic_pca_2_partition = TensorMap(
    '22009_Genetic-principal-components_0_2', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_2': 0},
)
genetic_pca_3_partition = TensorMap(
    '22009_Genetic-principal-components_0_3', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_3': 0},
)
genetic_pca_4_partition = TensorMap(
    '22009_Genetic-principal-components_0_4', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_4': 0},
)
genetic_pca_5_partition = TensorMap(
    '22009_Genetic-principal-components_0_5', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.014422761536727896, 'std': 10.57799283718005},
    days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_5': 0},
)
genetic_pca_6_partition = TensorMap(
    '22009_Genetic-principal-components_0_6', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_6': 0},
)

genetic_pca_7_partition = TensorMap(
    '22009_Genetic-principal-components_0_7', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_7': 0},
)

genetic_pca_8_partition = TensorMap(
    '22009_Genetic-principal-components_0_8', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_8': 0},
)

genetic_pca_9_partition = TensorMap(
    '22009_Genetic-principal-components_0_9', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_9': 0},
)

genetic_pca_10_partition = TensorMap(
    '22009_Genetic-principal-components_0_10', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_10': 0},
)

genetic_pca_11_partition = TensorMap(
    '22009_Genetic-principal-components_0_11', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_11': 0},
)

genetic_pca_12_partition = TensorMap(
    '22009_Genetic-principal-components_0_12', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_12': 0},
)

genetic_pca_13_partition = TensorMap(
    '22009_Genetic-principal-components_0_13', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_13': 0},
)

genetic_pca_14_partition = TensorMap(
    '22009_Genetic-principal-components_0_14', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_14': 0},
)

genetic_pca_15_partition = TensorMap(
    '22009_Genetic-principal-components_0_15', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_15': 0},
)

genetic_pca_16_partition = TensorMap(
    '22009_Genetic-principal-components_0_16', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_16': 0},
)

genetic_pca_17_partition = TensorMap(
    '22009_Genetic-principal-components_0_17', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_17': 0},
)

genetic_pca_18_partition = TensorMap(
    '22009_Genetic-principal-components_0_18', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_18': 0},
)

genetic_pca_19_partition = TensorMap(
    '22009_Genetic-principal-components_0_19', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_19': 0},
)

genetic_pca_20_partition = TensorMap(
    '22009_Genetic-principal-components_0_20', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.0144, 'std': 10.578}, days_window=offset, annotation_units=64,
    loss='log_cosh', channel_map={'22009_Genetic-principal-components_0_20': 0},
)

genetic_pca_all5 = TensorMap(
    'genetic_pca_all5', Interpretation.CONTINUOUS, path_prefix='continuous',
    normalization={'mean': -0.014, 'std': 10.578}, loss='log_cosh', annotation_units=5, shape=(5,), activation='linear',
    channel_map={
        '22009_Genetic-principal-components_0_0': 0, '22009_Genetic-principal-components_0_1': 1,
        '22009_Genetic-principal-components_0_2': 2, '22009_Genetic-principal-components_0_3': 3,
        '22009_Genetic-principal-components_0_4': 4,
    },
)

genetic_caucasian = TensorMap(
    'Genetic-ethnic-grouping_Caucasian_0_0', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_FLAG,
    channel_map={'no_caucasian': 0, 'Genetic-ethnic-grouping_Caucasian_0_0': 1},
)

genetic_caucasian_weighted = TensorMap(
    'Genetic-ethnic-grouping_Caucasian_0_0', Interpretation.CATEGORICAL, path_prefix='categorical',
    storage_type=StorageType.CATEGORICAL_FLAG,
    channel_map={'no_caucasian': 0, 'Genetic-ethnic-grouping_Caucasian_0_0': 1},
    loss=weighted_crossentropy([10.0, 1.0], 'caucasian_loss'),
)


def negative_mean_squared_error(y_true, y_pred):
    return -1 * mean_squared_error(y_true, y_pred)


negative_genetic_pca_1 = TensorMap(
    '22009_Genetic-principal-components_0_1', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=negative_mean_squared_error, activation='linear', channel_map={'22009_Genetic-principal-components_0_1': 0},
)

pearson_loss_genetic_pca_1 = TensorMap(
    'ploss_pca_1', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_1': 0},
)
pearson_loss_genetic_pca_2 = TensorMap(
    'ploss_pca_2', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_2': 0},
)
pearson_loss_genetic_pca_3 = TensorMap(
    'ploss_pca_3', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_3': 0},
)
pearson_loss_genetic_pca_4 = TensorMap(
    'ploss_pca_4', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_4': 0},
)
pearson_loss_genetic_pca_5 = TensorMap(
    'ploss_pca_5', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_5': 0},
)
pearson_loss_genetic_pca_6 = TensorMap(
    'ploss_pca_6', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_6': 0},
)
pearson_loss_genetic_pca_7 = TensorMap(
    'ploss_pca_7', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_7': 0},
)
pearson_loss_genetic_pca_8 = TensorMap(
    'ploss_pca_8', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_8': 0},
)
pearson_loss_genetic_pca_9 = TensorMap(
    'ploss_pca_9', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_9': 0},
)
pearson_loss_genetic_pca_10 = TensorMap(
    'ploss_pca_10', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_10': 0},
)
pearson_loss_genetic_pca_11 = TensorMap(
    'ploss_pca_11', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_11': 0},
)
pearson_loss_genetic_pca_12 = TensorMap(
    'ploss_pca_12', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_12': 0},
)
pearson_loss_genetic_pca_13 = TensorMap(
    'ploss_pca_13', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_13': 0},
)
pearson_loss_genetic_pca_14 = TensorMap(
    'ploss_pca_14', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_14': 0},
)
pearson_loss_genetic_pca_15 = TensorMap(
    'ploss_pca_15', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_15': 0},
)
pearson_loss_genetic_pca_16 = TensorMap(
    'ploss_pca_16', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_16': 0},
)
pearson_loss_genetic_pca_17 = TensorMap(
    'ploss_pca_17', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_17': 0},
)
pearson_loss_genetic_pca_18 = TensorMap(
    'ploss_pca_18', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_18': 0},
)
pearson_loss_genetic_pca_19 = TensorMap(
    'ploss_pca_19', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_19': 0},
)
pearson_loss_genetic_pca_20 = TensorMap(
    'ploss_pca_20', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_20': 0},
)
pearson_loss_genetic_pca_21 = TensorMap(
    'ploss_pca_21', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_21': 0},
)
pearson_loss_genetic_pca_22 = TensorMap(
    'ploss_pca_22', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_22': 0},
)
pearson_loss_genetic_pca_23 = TensorMap(
    'ploss_pca_23', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_23': 0},
)
pearson_loss_genetic_pca_24 = TensorMap(
    'ploss_pca_24', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_24': 0},
)
pearson_loss_genetic_pca_25 = TensorMap(
    'ploss_pca_25', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_25': 0},
)
pearson_loss_genetic_pca_26 = TensorMap(
    'ploss_pca_26', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_26': 0},
)
pearson_loss_genetic_pca_27 = TensorMap(
    'ploss_pca_27', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_27': 0},
)
pearson_loss_genetic_pca_28 = TensorMap(
    'ploss_pca_28', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_28': 0},
)
pearson_loss_genetic_pca_29 = TensorMap(
    'ploss_pca_29', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_29': 0},
)
pearson_loss_genetic_pca_30 = TensorMap(
    'ploss_pca_30', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_30': 0},
)
pearson_loss_genetic_pca_31 = TensorMap(
    'ploss_pca_31', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_31': 0},
)
pearson_loss_genetic_pca_32 = TensorMap(
    'ploss_pca_32', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_32': 0},
)
pearson_loss_genetic_pca_33 = TensorMap(
    'ploss_pca_33', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_33': 0},
)
pearson_loss_genetic_pca_34 = TensorMap(
    'ploss_pca_34', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_34': 0},
)
pearson_loss_genetic_pca_35 = TensorMap(
    'ploss_pca_35', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_35': 0},
)
pearson_loss_genetic_pca_36 = TensorMap(
    'ploss_pca_36', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_36': 0},
)
pearson_loss_genetic_pca_37 = TensorMap(
    'ploss_pca_37', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_37': 0},
)
pearson_loss_genetic_pca_38 = TensorMap(
    'ploss_pca_38', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_38': 0},
)
pearson_loss_genetic_pca_39 = TensorMap(
    'ploss_pca_39', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_39': 0},
)
pearson_loss_genetic_pca_40 = TensorMap(
    'ploss_pca_40', Interpretation.CONTINUOUS, path_prefix='continuous',
    loss=abs_pearson, metrics=[pearson], channel_map={'22009_Genetic-principal-components_0_40': 0},
)



import logging
import tempfile
from typing import Dict, List

import h5py
import apache_beam as beam
from apache_beam import Pipeline
from google.cloud import storage

from ml4h.defines import TENSOR_EXT, GCS_BUCKET, JOIN_CHAR, CONCAT_CHAR, HD5_GROUP_CHAR, dataset_name_from_meaning


def tensorize_sql_fields(pipeline: Pipeline, output_path: str, sql_dataset: str, tensor_type: str):

    if tensor_type == 'categorical':
        query = _get_categorical_query(sql_dataset)
    elif tensor_type == 'continuous':
        query = _get_continuous_query(sql_dataset)
    elif tensor_type == 'icd':
        query = _get_icd_query(sql_dataset)
    elif tensor_type == 'disease':
        query = _get_disease_query(sql_dataset)
    elif tensor_type == 'phecode_disease':
        query = _get_phecode_query(sql_dataset)
    elif tensor_type == 'death':
        query = _get_death_and_censor_query(sql_dataset)
    else:
        raise ValueError("Can tensorize only categorical or continuous fields, got ", tensor_type)

    bigquery_source = beam.io.BigQuerySource(query=query, use_standard_sql=True)
    # Query table in BQ
    steps = (
            pipeline
            | 'QueryTables' >> beam.io.Read(bigquery_source)

            # Each row is a dictionary where the keys are the BigQuery columns
            | 'CreateKey' >> beam.Map(lambda row: (row['sample_id'], row))

            # Group by key
            | 'GroupByKey' >> beam.GroupByKey()

            # Format into hd5 files and upload to GCS
            | 'CreateHd5sAndUploadToGCS' >> beam.Map(write_tensor_from_sql, output_path, tensor_type)
    )

    result = pipeline.run()
    result.wait_until_finish()


# We are instantiating the GCS client in global scope because passing it explicitly into a method used by beam.Map()
# gives a 'client not picklable` error. We're also enclosing it in a 'try' block so when this module is imported and if
# the client instantiation fails (when training models in Docker for instance), the whole run doesn't error out.
try:
    gcs_client = storage.Client()
    output_bucket = gcs_client.get_bucket(GCS_BUCKET)
# except OSError:
except:
    output_bucket ='nope'
    logging.warning(f"no GCS storage client")


def write_tensor_from_sql(sampleid_to_rows, output_path, tensor_type):
    # GroupByKey output is not a list of dicts, as expected, but something called an 'UnwindowedValue'.
    # We convert it to list first to arrive at (key, [{dict1}, {dict2}, {dict3}...])
    (sample_id, values) = sampleid_to_rows
    rows = list(values)

    try:
        with tempfile.TemporaryDirectory() as temp_dir:
            tensor_file = f"{sample_id}{TENSOR_EXT}"
            tensor_path = f"{temp_dir}/{tensor_file}"
            gcs_blob = output_bucket.blob(f"{output_path}/{tensor_file}")
            logging.info(f"Writing tensor {tensor_file} to {gcs_blob.public_url} ...")
            with h5py.File(tensor_path, 'w') as hd5:
                if tensor_type == 'icd':
                    icds = sorted(list(set([row['value'] for row in rows])))
                    hd5.create_dataset('icd', (1,), data=JOIN_CHAR.join(icds), dtype=h5py.special_dtype(vlen=str))
                elif tensor_type == 'categorical':
                    for row in rows:
                        fields = [
                            str(row['fieldid']), row['field'], row['meaning'],
                            str(row['instance']), str(row['array_idx']),
                        ]
                        hd5_dataset_name = dataset_name_from_meaning('categorical', fields)
                        _write_float_or_warn(sample_id, row, hd5_dataset_name, hd5)
                elif tensor_type == 'continuous':
                    for row in rows:
                        fields = [str(row['fieldid']), row['field'], str(row['instance']), str(row['array_idx'])]
                        hd5_dataset_name = dataset_name_from_meaning('continuous', fields)
                        _write_float_or_warn(sample_id, row, hd5_dataset_name, hd5)
                elif tensor_type in ['disease', 'phecode_disease']:
                    for row in rows:
                        hd5.create_dataset('categorical' + HD5_GROUP_CHAR + row['disease'].lower(), data=[float(row['has_disease'])])
                        hd5_date = 'dates' + HD5_GROUP_CHAR + row['disease'].lower() + '_date'
                        hd5.create_dataset(hd5_date, (1,), data=str(row['censor_date']), dtype=h5py.special_dtype(vlen=str))
                elif tensor_type == 'death':
                    for row in rows:
                        hd5.create_dataset('categorical' + HD5_GROUP_CHAR + 'death', data=[float(row['has_died'])])
                        d = 'dates' + HD5_GROUP_CHAR
                        hd5.create_dataset(d+'enroll_date', (1,), data=str(row['enroll_date']), dtype=h5py.special_dtype(vlen=str))
                        hd5.create_dataset(d+'death_censor', (1,), data=str(row['death_censor_date']), dtype=h5py.special_dtype(vlen=str))
                        hd5.create_dataset(d+'phenotype_censor', (1,), data=str(row['phenotype_censor_date']), dtype=h5py.special_dtype(vlen=str))
            gcs_blob.upload_from_filename(tensor_path)
    except:
        logging.exception(f"Problem with processing sample id '{sample_id}'")


def _write_float_or_warn(sample_id, row, hd5_dataset_name, hd5):
    try:
        float_value = float(row['value'])
        hd5.create_dataset(hd5_dataset_name, data=[float_value])
    except ValueError:
        logging.warning(f"Cannot cast to float from '{row['value']}' for field id '{row['fieldid']}' and sample id '{sample_id}'")


def _get_categorical_query(dataset):
    return f"""
        SELECT c.meaning, p_f_d.sample_id, p_f_d.fieldid, p_f_d.field, p_f_d.instance, p_f_d.array_idx, p_f_d.value
        FROM
        (
            SELECT f_d.field, p.sample_id, p.fieldid, p.instance, p.array_idx, p.value, p.coding_file_id
            FROM `{dataset}.phenotype` p
            INNER JOIN
            (
              SELECT d.fieldid, d.field
              FROM `shared_data.tensorization_fieldids` f
              INNER JOIN `{dataset}.dictionary` d
              ON f.fieldid = d.fieldid
              WHERE d.valuetype IN ('Categorical single', 'Categorical multiple')
            ) AS f_d
            ON f_d.fieldid = p.fieldid
        ) AS p_f_d
        INNER JOIN `{dataset}.coding` c
        ON c.coding=p_f_d.value AND c.coding_file_id=p_f_d.coding_file_id
    """


def _get_continuous_query(dataset):
    return f"""
        SELECT f_d.field, p.sample_id, p.fieldid, p.instance, p.array_idx, p.value
        FROM `{dataset}.phenotype` p
        INNER JOIN
        (
          SELECT d.fieldid, d.field
          FROM `shared_data.tensorization_fieldids` f
          INNER JOIN `{dataset}.dictionary` d
          ON f.fieldid = d.fieldid
          WHERE d.valuetype IN ('Continuous', 'Integer')
        ) AS f_d
        ON f_d.fieldid = p.fieldid
    """


def _get_icd_query(dataset):
    return f"""
        SELECT sample_id, value FROM `{dataset}.phenotype` WHERE fieldid IN (41202, 41204, 40001, 40002, 40006);
    """


def _get_disease_query(dataset):
    return f"""
        SELECT sample_id, disease, has_disease, censor_date FROM `{dataset}.disease` WHERE has_disease=1;
    """


def _get_death_and_censor_query(dataset):
    return f"""
        SELECT distinct(d.sample_id), d.has_died, d.death_censor_date, c.phenotype_censor_date, d.enroll_date
        FROM `{dataset}.disease` d INNER JOIN `{dataset}.censor` c
         ON c.sample_id = d.sample_id ORDER BY d.sample_id;
    """


def _get_phecode_query(dataset):
    return f"""
        SELECT sample_id, disease, has_disease, censor_date FROM `{dataset}.phecodes_nonzero` WHERE has_disease=1;
    """



import os
import pytest
import logging
import pandas as pd
import numpy as np

from ml4h.recipes import inference_file_name, _hidden_file_name,save_to_google_cloud
from ml4h.recipes import train_legacy, train_multimodal_multitask
from ml4h.recipes import infer_multimodal_multitask, infer_hidden_layer_multimodal_multitask
from ml4h.recipes import compare_multimodal_scalar_task_models, _find_learning_rate
from ml4h.explorations import _categorical_explore_header, _should_error_detect, explore
# Imports with test in their name
from ml4h.recipes import test_multimodal_multitask as tst_multimodal_multitask
from ml4h.recipes import test_multimodal_scalar_tasks as tst_multimodal_scalar_tasks
from ml4h.test_utils import TMAPS_UP_TO_4D
from ml4h.test_utils import build_hdf5s
from ml4h.TensorMap import TensorMap, Interpretation


class TestRecipes:
    def test_train(self, default_arguments):
        default_arguments.named_outputs = True
        train_multimodal_multitask(default_arguments)

    def test_train_legacy(self, default_arguments):
        default_arguments.named_outputs = True
        train_legacy(default_arguments)

    def test_test(self, default_arguments):
        tst_multimodal_multitask(default_arguments)

    def test_test_scalar(self, default_arguments):
        tst_multimodal_scalar_tasks(default_arguments)

    def test_infer(self, default_arguments):
        infer_multimodal_multitask(default_arguments)
        tsv = inference_file_name(default_arguments.output_folder, default_arguments.id)
        inferred = pd.read_csv(tsv, sep='\t')
        assert len(set(inferred['sample_id'])) == pytest.N_TENSORS

    def test_infer_genetics(self, default_arguments):
        default_arguments.tsv_style = 'genetics'
        infer_multimodal_multitask(default_arguments)
        tsv = inference_file_name(default_arguments.output_folder, default_arguments.id)
        inferred = pd.read_csv(tsv, sep='\t')
        default_arguments.tsv_style = 'standard'
        assert len(set(inferred['FID'])) == pytest.N_TENSORS

    def test_infer_hidden(self, default_arguments):
        infer_hidden_layer_multimodal_multitask(default_arguments)
        tsv = _hidden_file_name(default_arguments.output_folder, default_arguments.hidden_layer, default_arguments.id, '.tsv')
        inferred = pd.read_csv(tsv, sep='\t')
        assert len(set(inferred['sample_id'])) == pytest.N_TENSORS

    def test_find_learning_rate(self, default_arguments):
        default_arguments.named_outputs = True
        _find_learning_rate(default_arguments)

    @pytest.mark.slow
    def test_save_to_google_cloud(self,default_arguments):
        save_to_google_cloud(default_arguments)


    def test_explore(self, default_arguments, tmpdir_factory):
        temp_dir = tmpdir_factory.mktemp('explore_tensors2')
        default_arguments.tensors = str(temp_dir)
        tmaps = TMAPS_UP_TO_4D[:]
        tmaps.append(TensorMap(f'scalar', shape=(1,), interpretation=Interpretation.CONTINUOUS))
        explore_expected = build_hdf5s(temp_dir, tmaps, n=pytest.N_TENSORS, keys_are_paths=False)
        default_arguments.num_workers = 3
        default_arguments.tensor_maps_in = tmaps
        explore(default_arguments)

        csv_path = os.path.join(
            default_arguments.output_folder, default_arguments.id, 'tensors_all_union.csv',
        )
        explore_result = pd.read_csv(csv_path)
        logging.info(f'Tested explore {[c for c in explore_expected]}')
        logging.info(f'Tested explore_result {[c for c in explore_result["sample_id"]]}')
        for row in explore_result.iterrows():
            row = row[1]
            for tm in tmaps:
                row_expected = explore_expected[(row['sample_id'], tm.name)]
                if _should_error_detect(tm):
                    actual = getattr(row, tm.name)
                    assert not np.isnan(actual)
                    continue
                if tm.is_continuous():
                    actual = getattr(row, tm.name)
                    assert actual == row_expected
                    continue
                if tm.is_categorical():
                    for channel, idx in tm.channel_map.items():
                        channel_val = getattr(row, _categorical_explore_header(tm, channel))
                        assert channel_val == row_expected[idx]



import os
import csv
import logging
import operator

import h5py
import numpy as np
import pandas as pd
from typing.io import TextIO
from typing import List, Tuple

from ml4h.metrics import sparse_cross_entropy
from ml4h.TensorMap import TensorMap, Interpretation
from ml4h.DatabaseClient import BigQueryDatabaseClient, DatabaseClient
from ml4h.defines import TENSOR_MAPS_FILE_NAME, dataset_name_from_meaning
from ml4h.defines import DICTIONARY_TABLE, CODING_TABLE, PHENOTYPE_TABLE, JOIN_CHAR
from ml4h.tensormap.general import build_tensor_from_file, build_categorical_tensor_from_file
from ml4h.tensormap.text import random_text_window_tensor, token_dictionary_and_text_from_file, token_dictionary_from_hd5_key, random_array_window_tensors
from ml4h.tensorize.tensor_writer_ukbb import disease_prevalence_status, get_disease2tsv, disease_incidence_status, disease_censor_status


LESS_THAN_CODES = "('Less than a year', 'Less than once a week', 'Less than one mile', 'Less than an hour a day', 'Less than one a day', 'Less than one', 'Less than once a year', 'Less than 1 year ago', 'Less than a year ago', 'Less than one year', 'Less than one cigarette per day')"


def make_test_tensor_maps(desired_map_name: str) -> TensorMap:
    for n in range(1, 6):
        if desired_map_name == f'{n}d_cont':
            return TensorMap(f'{n}d_cont', shape=tuple(range(2, n + 2)), interpretation=Interpretation.CONTINUOUS)
        if desired_map_name == f'{n}d_cat':
            return TensorMap(
                f'{n}d_cat', shape=tuple(range(2, n + 2)), interpretation=Interpretation.CATEGORICAL,
                channel_map={f'c_{i}': i for i in range(n + 1)},
            )


def write_tensor_maps(args) -> None:
    logging.info("Making tensor maps...")

    tensor_maps_file = f"{args.output_folder}/{TENSOR_MAPS_FILE_NAME}.py"
    db_client = BigQueryDatabaseClient(credentials_file=args.bigquery_credentials_file)
    with open(tensor_maps_file, 'w') as f:
        f.write(_get_tensor_map_file_imports())
        #_write_disease_tensor_maps(args.phenos_folder, f)
        #_write_disease_tensor_maps_incident_prevalent(args.phenos_folder, f)
        #_write_phecode_tensor_maps(f, args.phecode_definitions, db_client)
        _write_continuous_tensor_maps(f, db_client)

        f.write('\n')
        logging.info(f"Wrote the tensor maps to {tensor_maps_file}.")


def _get_tensor_map_file_imports() -> str:
    return (
        f"# TensorMaps automatically generated by {os.path.basename(__file__)}\n"
        f"# DO NOT EDIT\n\n"
        f"from ml4h.defines import StorageType\n"
        f"from ml4h.metrics import weighted_crossentropy\n"
        f"from ml4h.tensormap.ukb.demographics import prevalent_incident_tensor\n"
        f"from ml4h.TensorMap import TensorMap, Interpretation, make_range_validator\n\n"
    )


def _write_disease_tensor_maps(phenos_folder: str, f: TextIO) -> None:
    f.write(f"\n\n#  TensorMaps for MPG disease phenotypes\n")
    disease2tsv = get_disease2tsv(phenos_folder)
    logging.info(f"Got:{len(disease2tsv)} disease TSVs from:{phenos_folder}")
    status = disease_censor_status(disease2tsv, 1000000, 5000000)
    logging.info(f"Got status for all diseases.")
    for d in sorted(list(disease2tsv.keys())):
        total = len(status[d])
        diseased = np.sum(list(status[d].values()))
        factor = int(total / (1 + diseased * 2))
        f.write(
            f"{d} = TensorMap('{d}', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical', "
            f"channel_map={{'no_{d}':0, '{d}':1}}, loss=weighted_crossentropy([1.0, {factor}], '{d}'))\n",
        )
    logging.info(f"Done writing TensorMaps for diseases.")


def _write_disease_tensor_maps_incident_prevalent(phenos_folder: str, f: TextIO) -> None:
    f.write(f"\n\n#  TensorMaps for prevalent and incident MPG disease phenotypes\n")
    disease2tsv = get_disease2tsv(phenos_folder)
    logging.info(f"Got:{len(disease2tsv)} disease TSVs from:{phenos_folder}")
    status_p = disease_prevalence_status(disease2tsv, 1000000, 2500000)
    status_i = disease_incidence_status(disease2tsv, 1000000, 2500000)
    logging.info(f"Got prevalence and incidence status for all diseases.")
    for disease in sorted(list(disease2tsv.keys())):
        total = len(status_p[disease])
        diseased_p = np.sum(list(status_p[disease].values()))
        factor_p = int(total / (1 + (diseased_p * 3)))
        diseased_i = np.sum(list(status_i[disease].values()))
        factor_i = int(total / (1 + (diseased_i * 3)))
        f.write(
            f"{disease}_prevalent_incident = TensorMap('{disease}', Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, "
            f"path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/{disease}_date'), "
            f"channel_map={{'no_{disease}':0, 'prevalent_{disease}':1, 'incident_{disease}':2}}, "
            f"loss=weighted_crossentropy([1.0, {factor_p}, {factor_i}], '{disease}_prevalent_incident'))\n",
        )
    logging.info(f"Done writing TensorMaps for prevalent and incident diseases.")


def _write_phecode_tensor_maps(f: TextIO, phecode_csv, db_client: DatabaseClient):
    # phecode_csv = '/home/sam/phecode_definitions1.2.csv'
    total_samples = 500000
    remove_chars = ";.,/()-[]&' "
    phecode2phenos = {}
    with open(phecode_csv, 'r') as my_csv:
        lol = list(csv.reader(my_csv, delimiter=','))
        for row in lol[1:]:
            pheno = row[1].strip().replace("'s", "s")
            for c in remove_chars:
                pheno = pheno.replace(c, '_')
            pheno = pheno.lower().strip('_').replace('___', '_').replace('__', '_')
            phecode2phenos['phecode_'+row[0].lstrip('0').strip()] = pheno
    query = f"select disease, count(disease) as total from `broad-ml4cvd.ukbb7089_201904.phecodes_nonzero` GROUP BY disease"
    count_result = db_client.execute(query)
    phecode2counts = {}
    for row in count_result:
        phecode2counts[row['disease']] = float(row['total'])

    f.write(f"\n\n#  TensorMaps for Phecode disease phenotypes\n")
    for k, p in sorted(phecode2phenos.items(), key=operator.itemgetter(1)):
        if k in phecode2counts:
            factor = int(total_samples / (1+phecode2counts[k]))
            f.write(
                f"{p}_phe = TensorMap('{k}', Interpretation.CATEGORICAL, channel_map={{'no_{p}':0, '{p}':1}}, path_prefix='categorical', "
                f"storage_type=StorageType.CATEGORICAL_FLAG, loss=weighted_crossentropy([1.0, {factor}], '{k.replace('.', '_')}'))\n",
            )

    query = f"select disease, count(disease) as total from `broad-ml4cvd.ukbb7089_201904.phecodes_nonzero` WHERE prevalent_disease=1 GROUP BY disease"
    count_result = db_client.execute(query)
    phecode2prevalent = {}
    for row in count_result:
        phecode2prevalent[row['disease']] = float(row['total'])

    query = f"select disease, count(disease) as total from `broad-ml4cvd.ukbb7089_201904.phecodes_nonzero` WHERE incident_disease=1 GROUP BY disease"
    count_result = db_client.execute(query)
    phecode2incident = {}
    for row in count_result:
        phecode2incident[row['disease']] = float(row['total'])

    f.write(f"\n\n#  TensorMaps for prevalent and incident Phecode disease phenotypes\n")
    for k, p in sorted(phecode2phenos.items(), key=operator.itemgetter(1)):
        if k in phecode2incident and k in phecode2prevalent:
            factor_i = int(total_samples / (1 + phecode2incident[k]))
            factor_p = int(total_samples / (1 + phecode2prevalent[k]))
            f.write(
                f"{p}_phe_pi = TensorMap('{k}',  Interpretation.CATEGORICAL,  storage_type=StorageType.CATEGORICAL_FLAG, "
                f"path_prefix='categorical', tensor_from_file=prevalent_incident_tensor('dates/enroll_date', 'dates/{k}_date'), "
                f"channel_map={{'no_{p}':0, '{p}_prevalent':1, '{p}_incident':2}}, "
                f"loss=weighted_crossentropy([1.0, {factor_p}, {factor_i}], '{p}_pi'))\n",
            )


def _write_continuous_tensor_maps(f: TextIO, db_client: DatabaseClient):
    # Handle special coding values in continuous variables in order to generate summary statistics (mean and std dev) for
    # each field across all samples. This will remove missing samples from the calculation and change the value of 'Less than one'
    query = f"""
    WITH coding_tmp AS (
        SELECT
            *,
            CASE
                WHEN meaning IN ('Do not know',  'Prefer not to answer', 'Ongoing when data entered') OR meaning LIKE "Still taking%" THEN TRUE
            END AS missing,
            CASE
                WHEN meaning IN {LESS_THAN_CODES} THEN '.5'
            END AS value
        FROM
            {CODING_TABLE}
    ), pheno_tmp AS (
    SELECT
        sample_id,
        FieldID,
        instance,
        array_idx,
        COALESCE(c.value, p.value) new_value,
        COALESCE(c.missing, FALSE) missing
    FROM {PHENOTYPE_TABLE} AS p
    LEFT JOIN coding_tmp AS c
        ON TRUE
        AND SAFE_CAST(p.value AS FLOAT64) = SAFE_CAST(c.coding AS FLOAT64)
        AND p.coding_file_id = c.coding_file_id
    )

    SELECT
        t.FieldID,
        Field,
        t.instance,
        AVG(CAST(new_value AS FLOAT64)) mean,
        STDDEV(CAST(new_value AS FLOAT64)) std,
        MAX(array_idx) AS max_array
    FROM pheno_tmp AS t
    LEFT JOIN {DICTIONARY_TABLE} AS d ON d.FieldID = t.FieldID
    WHERE TRUE
        AND ValueType IN ('Integer', 'Continuous')
        AND NOT missing
    GROUP BY t.FieldID, t.instance, Field ORDER BY t.FieldID
    """

    field_data_for_tensor_maps = db_client.execute(query)

    f.write(f"\n\n#  Continuous tensor maps\n")
    for row in field_data_for_tensor_maps:
        name = dataset_name_from_meaning(None, [str(row.FieldID), row.Field, str(row.instance)])
        channel_map = "channel_map={"
        for i in range(0, row.max_array + 1):
            channel_map += f"'{name}{JOIN_CHAR}{i}': {i}, "
        channel_map += "}"
        f.write(f"ukb_{row.FieldID}_{row.instance} = TensorMap('{name}{JOIN_CHAR}{i}', loss='log_cosh', path_prefix='continuous', ")
        f.write(f"normalization={{'mean': {row.mean}, 'std': {row.std}}}, {channel_map})\n")


def _get_pkl_path_for_field(field_id: int, pyukbb_data_path: str):
    """Returns the path to the .pkl file that contained `UKBioBankParsedField` for a given FieldID."""
    for _, _, files in os.walk(pyukbb_data_path):
        for file in files:
            if file.find(f'FieldID_{field_id}.pkl') > -1:
                return os.path.join(pyukbb_data_path, file)
    raise FileNotFoundError('Cannot find pyukbb .pkl file for field ID {field_id}!')


def _get_all_available_fields(available_fields_pd, keyword: str = None, category: int = None):
    filtered = available_fields_pd
    if category is not None:
        filtered = filtered[filtered.Category == category]
    if keyword is not None:
        filtered = filtered[filtered.Field.str.contains(keyword, case=False)]
    return filtered


def generate_continuous_tensor_map_from_file(
    file_name: str,
    column_name,
    tensor_map_name: str,
    normalization: bool,
    discretization_bounds: List[float],
) -> TensorMap:
    if discretization_bounds:
        return TensorMap(
            f'{tensor_map_name}', Interpretation.DISCRETIZED, channel_map={tensor_map_name: 0},
            tensor_from_file=build_tensor_from_file(file_name, column_name, normalization),
            discretization_bounds=discretization_bounds,
        )
    else:
        return TensorMap(
            f'{tensor_map_name}', channel_map={tensor_map_name: 0},
            annotation_units=64, days_window=0,
            tensor_from_file=build_tensor_from_file(file_name, column_name, normalization),
        )


def generate_categorical_tensor_map_from_file(
    file_name: str,
    column_name: str,
    tensor_map_name: str,
) -> TensorMap:
    ext = file_name.split('.')[1]
    delimiter = ',' if ext == 'csv' else '\t'
    df = pd.read_csv(file_name, delimiter=delimiter)
    channel_map = {}
    for i, k in enumerate(df[column_name].value_counts().keys()):
        channel_map[k] = i
    return TensorMap(
            f'{tensor_map_name}', Interpretation.CATEGORICAL, channel_map=channel_map,
            tensor_from_file=build_categorical_tensor_from_file(file_name, column_name),
    )


def _space_tensor_from_file(df: pd.DataFrame, dimensions: int, sample_column: str = 'sample_id'):
    def tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents=None):
        sample_id = int(os.path.basename(hd5.filename).replace('.hd5', ''))
        row = df[df[sample_column] == sample_id]
        if len(row) == 0:
            raise KeyError(f'Sample id not in dataframe.')
        values = row.iloc[0].tolist()[1:dimensions+1]
        return np.array(values, dtype=np.float32)
    return tensor_from_file


def generate_latent_tensor_map_from_file(
    file_name: str,
    tensor_map_name: str,
) -> TensorMap:
    delimiter = ',' if file_name.split('.')[1].lower() == 'csv' else '\t'
    df = pd.read_csv(file_name, delimiter=delimiter)
    dimensions = len(df.columns)-1
    return TensorMap(
            f'{tensor_map_name}', Interpretation.CONTINUOUS, shape=(dimensions,),
            tensor_from_file=_space_tensor_from_file(df, dimensions), metrics=[],
    )


def generate_random_text_tensor_maps(text_file: str, window_size: int) -> Tuple[TensorMap, TensorMap]:
    name = os.path.basename(text_file).split('.')[0]
    text, token_dictionary = token_dictionary_and_text_from_file(text_file)
    shape = (window_size,)

    output_map = TensorMap(
        f'next_{name}', Interpretation.LANGUAGE,
        shape=shape,
        loss=sparse_cross_entropy(window_size),
        channel_map=token_dictionary,
        cacheable=False,
    )
    input_map = TensorMap(
        name, Interpretation.LANGUAGE, shape=shape,
        tensor_from_file=random_text_window_tensor(text, window_size),
        dependent_map=output_map,
        channel_map=token_dictionary,
        cacheable=False,
    )
    return input_map, output_map


def generate_random_pixel_as_text_tensor_maps(
    tensors: str,
    path_prefix: str,
    window_shape: Tuple[int],
) -> Tuple[TensorMap, TensorMap]:
    name = path_prefix.split('/')[-1]
    path_prefix = '/'.join(path_prefix.split('/')[:-1])
    logging.info(f'prefix {path_prefix} and name is {name}')
    token_dictionary = token_dictionary_from_hd5_key(tensors, path_prefix, name)
    window_size = int(np.prod(window_shape))
    shape = (window_size,)
    output_map = TensorMap(
        f'next_{name}', Interpretation.LANGUAGE,
        shape=shape,
        loss=sparse_cross_entropy(window_size),
        channel_map=token_dictionary,
        cacheable=False,
    )
    input_map = TensorMap(
        name, Interpretation.LANGUAGE, shape=shape, path_prefix=path_prefix,
        tensor_from_file=random_array_window_tensors(window_shape),
        dependent_map=output_map,
        channel_map=token_dictionary,
        cacheable=False,
    )
    return input_map, output_map



import numpy as np
from abc import ABC, abstractmethod

from ml4h.defines import EPS
from tensorflow.keras.applications import imagenet_utils


class Normalizer(ABC):
    @abstractmethod
    def normalize(self, tensor: np.ndarray) -> np.ndarray:
        """Shape preserving transformation"""
        pass

    def normalize_loading_option(self, tensor: np.ndarray, _) -> np.ndarray:
        """Shape preserving transformation for use with DataDescription.
         Defaults to the normalize function if not defined in the descendant"""
        return self.normalize(tensor)

    def un_normalize(self, tensor: np.ndarray) -> np.ndarray:
        """The inverse of normalize if possible. Otherwise identity."""
        return tensor


class Standardize(Normalizer):
    def __init__(self, mean: float, std: float):
        self.mean, self.std = mean, std

    def normalize(self, tensor: np.ndarray) -> np.ndarray:
        return (tensor - self.mean) / self.std

    def un_normalize(self, tensor: np.ndarray) -> np.ndarray:
        return tensor * self.std + self.mean


class ZeroMeanStd1(Normalizer):
    def normalize(self, tensor: np.ndarray) -> np.ndarray:
        tensor -= np.mean(tensor)
        tensor /= np.std(tensor) + EPS
        return tensor


class NonZeroNormalize(Normalizer):
    def normalize(self, tensor: np.ndarray) -> np.ndarray:
        nonzero = tensor > 0
        tensor[nonzero] = (tensor[nonzero] - tensor[nonzero].mean() + 1e-9) / (
            tensor[nonzero].std() + 1e-9
        )
        return tensor


class TopKNormalize(Normalizer):
    def __init__(self, n_top: int = 50):
        self.n_top = n_top

    def normalize(self, tensor: np.ndarray) -> np.ndarray:
        """Find top K itensity voxels are set upper range to the mean of those"""
        upper = np.mean(sorted(np.max(tensor, axis=-1).flatten())[::-1][0:self.n_top])
        tensor = np.where(tensor >= upper, upper, tensor)
        tensor /= tensor.max()
        return tensor


class ImagenetNormalizeTorch(Normalizer):
    def normalize(self, tensor: np.ndarray) -> np.ndarray:
        # This is equivalent to:
        # x /= 255.
        # mean = [0.485, 0.456, 0.406]
        # std = [0.229, 0.224, 0.225]
        # when mode is torch
        return imagenet_utils.preprocess_input(tensor, data_format=None, mode="torch")


class RandomStandardize(Normalizer):
    def __init__(self, mean: float, std: float, ratio: float = 0.5):
        self.mean, self.std, self.ratio = mean, std, ratio

    def normalize(self, tensor: np.ndarray) -> np.ndarray:
        if np.random.rand() > self.ratio:
            return (tensor - self.mean) / (self.std + EPS)
        else:
            return (tensor - np.mean(tensor)) / (np.std(tensor) + EPS)

    def un_normalize(self, tensor: np.ndarray) -> np.ndarray:
        return tensor * self.std + self.mean




# Imports
import time
import logging
import numpy as np
from typing import Iterable

import tensorflow.keras.backend as K
from tensorflow.keras.models import Model
from tensorflow.keras.utils import model_to_dot


def plot_and_time_model(
    model: Model,
    generate_train: Iterable,
    generate_valid: Iterable,
    batch_size: int,
    training_steps: int,
    inspect_show_labels: bool,
    image_path: str,
) -> Model:
    """Collect statistics on model inference and training times.

    Arguments:
        model: the model to inspect
        generate_train: training data generator function
        generate_valid: Validation data generator function
        batch_size: size of the mini-batches
        training_steps: number of optimization steps to take
        inspect_show_labels: if True, show layer labels on the architecture diagram
        image_path: file path of the architecture diagram

    Returns:
        The slightly optimized keras model
    """
    if image_path:
        _plot_dot_model_in_color(model_to_dot(model, show_shapes=inspect_show_labels, expand_nested=True, show_layer_activations=True), image_path, inspect_show_labels)
    t0 = time.time()
    _ = model.fit(generate_train, steps_per_epoch=training_steps, validation_steps=1, validation_data=generate_valid)
    t1 = time.time()
    n = batch_size * training_steps
    train_speed = (t1 - t0) / n
    logging.info(f'Spent:{(t1 - t0):0.2f} seconds training, Samples trained on:{n} Per sample training speed:{train_speed:0.3f} seconds.')

    t0 = time.time()
    inference_steps = max(1, training_steps//8)
    _ = model.predict(generate_valid, steps=inference_steps, verbose=1)
    t1 = time.time()
    inference_speed = (t1 - t0) / (batch_size * inference_steps)

    logging.info(f'Spent:{(t1 - t0):0.2f} seconds predicting, Samples inferred:{n} Per sample inference speed:{inference_speed:0.4f} seconds.')
    return model


def _plot_dot_model_in_color(dot, image_path, inspect_show_labels):
    import pydot
    legend = {}
    #logging.info(f'Got label {dot.get_subgraph_list()} \nFull node n{dot.get_node_list()}\n\n')
    for sg in dot.get_subgraph_list():
        add_labels(sg.get_nodes())
    add_labels(dot.get_nodes())
    for label in legend:
        legend_node = pydot.Node('legend' + label, label=label, shape="box", fillcolor=legend[label])
        dot.add_node(legend_node)

    logging.info(f'Saving architecture diagram to:{image_path}')
    dot.write_png(image_path)


def add_labels(nodes):
    for n in nodes:
        if n.get_label():
            #logging.info(f'Got label {n.get_label()} \nFull node n{n}\n\n')
            if 'Conv1' in n.get_label():
              #  legend['Conv1'] = "cyan"
                n.set_fillcolor("cyan")
            elif 'Conv2' in n.get_label():
             #   legend['Conv2'] = "deepskyblue1"
                n.set_fillcolor("deepskyblue1")
            elif 'Conv3' in n.get_label():
             #   legend['Conv3'] = "deepskyblue3"
                n.set_fillcolor("deepskyblue3")
            elif 'UpSampling' in n.get_label():
             #   legend['UpSampling'] = "darkslategray2"
                n.set_fillcolor("darkslategray2")
            elif 'Transpose' in n.get_label():
             #   legend['Transpose'] = "deepskyblue2"
                n.set_fillcolor("deepskyblue2")
            elif 'BatchNormalization' in n.get_label():
              #  legend['BatchNormalization'] = "goldenrod1"
                n.set_fillcolor("goldenrod1")
            elif 'output_' in n.get_label():
                n.set_fillcolor("darkolivegreen2")
             #   legend['Output'] = "darkolivegreen2"
            elif 'softmax' in n.get_label():
                n.set_fillcolor("chartreuse")
             #   legend['softmax'] = "chartreuse"
            elif 'MaxPooling' in n.get_label():
              #  legend['MaxPooling'] = "aquamarine"
                n.set_fillcolor("aquamarine")
            elif 'Dense' in n.get_label():
              #  legend['Dense'] = "gold"
                n.set_fillcolor("gold")
            elif 'Reshape' in n.get_label():
               # legend['Reshape'] = "coral"
                n.set_fillcolor("coral")
            elif 'Input' in n.get_label():
              #  legend['Input'] = "darkolivegreen1"
                n.set_fillcolor("darkolivegreen1")
            elif 'Activation' in n.get_label():
              #  legend['Activation'] = "yellow"
                n.set_fillcolor("yellow")
        n.set_style("filled")


def saliency_map(input_tensor: np.ndarray, model: Model, output_layer_name: str, output_index: int) -> np.ndarray:
    """Compute saliency maps of the given model (presumably already trained) on a batch of inputs with respect to the desired output layer and index.

    For example, with a trinary classification layer called quality and classes good, medium, and bad output layer name would be "quality_output"
    and output_index would be 0 to get gradients w.r.t. good, 1 to get gradients w.r.t. medium, and 2 for gradients w.r.t. bad.

    :param input_tensor: A batch of input tensors
    :param model: A trained model expecting those inputs
    :param output_layer_name: The name of the output layer that the derivative will be taken with respect to
    :param output_index: The index within the output layer that the derivative will be taken with respect to

    :return: Array of the gradients same shape as input_tensor
    """
    get_gradients = _gradients_from_output(model, output_layer_name, output_index)
    activation, gradients = get_gradients([input_tensor])
    return gradients


def _gradients_from_output(model, output_layer, output_index):
    K.set_learning_phase(1)
    input_tensor = model.input
    x = model.get_layer(output_layer).output[:, output_index]
    grads = K.gradients(x, input_tensor)[0]
    grads /= (K.sqrt(K.mean(K.square(grads))) + 1e-6)  # normalization trick: we normalize the gradient
    iterate = K.function([input_tensor], [x, grads])
    return iterate



import sys
import pytest
from ml4h.arguments import parse_args
from ml4h.test_utils import TMAPS as MOCK_TMAPS


class TestUConnect:

    def test_no_u(self, tmpdir):
        sys.argv = [
            'train',
            '--output_folder', str(tmpdir),
        ]
        args = parse_args()
        assert len(args.u_connect) == 0

    def test_simple_u(self, tmpdir):
        inp_key = '3d_cont'
        sys.argv = [
            'train',
            '--output_folder', str(tmpdir),
            '--input_tensors', inp_key,
            '--output_tensors', inp_key,
            '--u_connect', inp_key, inp_key,
        ]
        args = parse_args()
        assert len(args.u_connect) == 1
        inp, out = list(args.u_connect.items())[0]
        tmap = MOCK_TMAPS[inp_key]
        assert inp == tmap
        assert out == {tmap}

    def test_many_to_one(self, tmpdir):
        inp_key1 = '3d_cont'
        inp_key2 = '3d_cat'
        sys.argv = [
            'train',
            '--output_folder', str(tmpdir),
            '--input_tensors', inp_key1, inp_key2,
            '--output_tensors', inp_key1,
            '--u_connect', inp_key1, inp_key1,
            '--u_connect', inp_key2, inp_key1,
        ]
        args = parse_args()
        assert len(args.u_connect) == 2
        assert args.u_connect[MOCK_TMAPS[inp_key1]] == {MOCK_TMAPS[inp_key1]}
        assert args.u_connect[MOCK_TMAPS[inp_key2]] == {MOCK_TMAPS[inp_key1]}

    def test_one_to_many(self, tmpdir):
        key1 = '3d_cont'
        key2 = '3d_cat'
        sys.argv = [
            'train',
            '--output_folder', str(tmpdir),
            '--input_tensors', key1, key2,
            '--output_tensors', key1, key2,
            '--u_connect', key1, key1,
            '--u_connect', key1, key2,
        ]
        args = parse_args()
        assert len(args.u_connect) == 1
        assert args.u_connect[MOCK_TMAPS[key1]] == {MOCK_TMAPS[key1], MOCK_TMAPS[key2]}

    def test_multi_u(self, tmpdir):
        key1 = '3d_cont'
        key2 = '3d_cat'
        sys.argv = [
            'train',
            '--output_folder', str(tmpdir),
            '--input_tensors', key1, key2,
            '--output_tensors', key1, key2,
            '--u_connect', key1, key1,
            '--u_connect', key2, key2,
        ]
        args = parse_args()
        assert len(args.u_connect) == 2
        assert args.u_connect[MOCK_TMAPS[key1]] == {MOCK_TMAPS[key1]}
        assert args.u_connect[MOCK_TMAPS[key2]] == {MOCK_TMAPS[key2]}



import logging
from typing import Dict, Tuple

import numpy as np
from tensorflow.keras.utils import to_categorical

from ml4h.TensorMap import TensorMap, Interpretation
from ml4h.defines import LAX_4CH_HEART_LABELS, LAX_4CH_MYOCARDIUM_LABELS, MRI_LAX_4CH_SEGMENTED_CHANNEL_MAP, ECG_REST_MEDIAN_LEADS
from ml4h.normalizer import ZeroMeanStd1, Standardize
from ml4h.tensormap.general import get_tensor_at_first_date, pad_or_crop_array_to_shape


def _make_ecg_rest(
        hd5, path_prefix: str, ecg_shape: Tuple[int], ecg_leads: Dict[str, int], instance: int = 2, downsample_steps: int = 0,
):
    tensor = np.zeros(ecg_shape, dtype=np.float32)
    for k in hd5[path_prefix]:
        if k in ecg_leads:
            data = hd5[f'{path_prefix}/{k}/instance_{instance}']
            if downsample_steps > 1:
                tensor[:, ecg_leads[k]] = np.array(data, dtype=np.float32)[::downsample_steps]
            tensor[:, ecg_leads[k]] = pad_or_crop_array_to_shape((ecg_shape[0],), data)
    return tensor


def _heart_mask_and_ecg_instances(
    mri_path_prefix, mri_shape, mri_key, mri_segmentation_key, mri_labels, mri_normalizer,
    ecg_prefix, ecg_shape, ecg_leads, ecg_normalizer, include_mri_segmentation=False, total_instances=50,
):
    def _heart_mask_tensor_from_file(tm, hd5, dependents={}):
        diastole_categorical = get_tensor_at_first_date(hd5, mri_path_prefix, f'{mri_segmentation_key}{1}')
        heart_mask = np.isin(diastole_categorical, list(mri_labels.values()))
        i, j = np.where(heart_mask)
        indices = np.meshgrid(np.arange(min(i), max(i) + 1), np.arange(min(j), max(j) + 1), np.arange(total_instances), indexing='ij')
        segmentation_indices = np.meshgrid(np.arange(min(i), max(i) + 1), np.arange(min(j), max(j) + 1), indexing='ij')
        mri = mri_normalizer.normalize(get_tensor_at_first_date(hd5, mri_path_prefix, f'{mri_key}'))
        if ecg_shape[0] > 0:
            ecg = ecg_normalizer.normalize(_make_ecg_rest(hd5, ecg_prefix, ecg_shape, ecg_leads))
        tensor = np.zeros(tm.shape, dtype=np.float32)
        for frame in range(1, total_instances+1):
            if include_mri_segmentation:
                frame_categorical = get_tensor_at_first_date(hd5, mri_path_prefix, f'{mri_segmentation_key}{frame}')
                heart_mask = np.isin(frame_categorical, list(mri_labels.values()))
                mri[:mri.shape[0], :mri.shape[1], frame-1] = heart_mask[:mri.shape[0], :mri.shape[1]] * mri[..., frame-1]
                frame_categorical = get_tensor_at_first_date(hd5, mri_path_prefix, f'{mri_segmentation_key}{frame}')
                reshape_categorical = pad_or_crop_array_to_shape(mri_shape[:2], frame_categorical[segmentation_indices])
                if len(mri_shape) == 4:
                    slice_one_hot = to_categorical(reshape_categorical, tm.shape[-1]-1)
                    tensor[:mri_shape[0], :mri_shape[1], frame - 1, 1:] = slice_one_hot
                else:
                    tensor[mri_shape[0]:, :mri_shape[1], frame - 1] = reshape_categorical
            if ecg_shape[0] > 0:
                ecg_start = (frame-1) * (ecg_shape[0] // total_instances)
                ecg_stop = frame * (ecg_shape[0] // total_instances)
                for lead in ecg_leads:
                    lead_index = ecg_leads[lead] + mri_shape[0]
                    if len(mri_shape) == 3:
                        tensor[lead_index, :, frame-1] = np.repeat(ecg[ecg_start:ecg_stop, ecg_leads[lead]], tm.shape[1]//(ecg_stop-ecg_start))
                    elif len(mri_shape) == 4:
                        tensor[lead_index, :, frame - 1, :] = np.expand_dims(np.repeat(ecg[ecg_start:ecg_stop, ecg_leads[lead]], tm.shape[1] // (ecg_stop - ecg_start)), axis=-1)
        if len(mri_shape) == 3:
            tensor[:mri_shape[0], :mri_shape[1], :mri_shape[2]] = pad_or_crop_array_to_shape(mri_shape, mri[tuple(indices)])
        elif len(mri_shape) == 4:
            tensor[:mri_shape[0], :mri_shape[1], :mri_shape[2], 0] = pad_or_crop_array_to_shape(mri_shape[:3], mri[tuple(indices)])
        return tensor
    return _heart_mask_tensor_from_file


tff = _heart_mask_and_ecg_instances(
    'ukb_cardiac_mri', (96, 96, 50), 'cine_segmented_lax_4ch/2/',
    'cine_segmented_lax_4ch_annotated_', LAX_4CH_HEART_LABELS, ZeroMeanStd1(),
    'ukb_ecg_rest', (600, 12), ECG_REST_MEDIAN_LEADS, Standardize(mean=0, std=10),
)
ecg_and_lax_4ch = TensorMap(
    'ecg_and_lax_4ch', Interpretation.CONTINUOUS, shape=(108, 96, 50),
    tensor_from_file=tff,
)
tff = _heart_mask_and_ecg_instances(
    'ukb_cardiac_mri', (92, 96, 50), 'cine_segmented_lax_4ch/2/',
    'cine_segmented_lax_4ch_annotated_', LAX_4CH_HEART_LABELS, ZeroMeanStd1(),
    'ukb_ecg_rest', (600, 12), ECG_REST_MEDIAN_LEADS, Standardize(mean=0, std=10),
)
ecg_and_lax_4ch_104 = TensorMap(
    'ecg_and_lax_4ch_104', Interpretation.CONTINUOUS, shape=(104, 96, 50),
    tensor_from_file=tff,
)

tff = _heart_mask_and_ecg_instances(
    'ukb_cardiac_mri', (96, 96, 50, len(MRI_LAX_4CH_SEGMENTED_CHANNEL_MAP)+1), 'cine_segmented_lax_4ch/2/',
    'cine_segmented_lax_4ch_annotated_', LAX_4CH_HEART_LABELS, ZeroMeanStd1(),
    'ukb_ecg_rest', (600, 12), ECG_REST_MEDIAN_LEADS, Standardize(mean=0, std=10),
    include_mri_segmentation=True,
)
ecg_seg_lax_4ch = TensorMap(
    'ecg_seg_lax_4ch', Interpretation.CONTINUOUS, shape=(108, 96, 50, len(MRI_LAX_4CH_SEGMENTED_CHANNEL_MAP)+1),
    tensor_from_file=tff, channel_map=MRI_LAX_4CH_SEGMENTED_CHANNEL_MAP,
)

tff = _heart_mask_and_ecg_instances(
    'ukb_cardiac_mri', (64, 64, 50, len(MRI_LAX_4CH_SEGMENTED_CHANNEL_MAP)+1), 'cine_segmented_lax_4ch/2/',
    'cine_segmented_lax_4ch_annotated_', LAX_4CH_HEART_LABELS, ZeroMeanStd1(),
    'ukb_ecg_rest', (0, 0), ECG_REST_MEDIAN_LEADS, Standardize(mean=0, std=10),
    include_mri_segmentation=True,
)
seg_lax_4ch_64xy_50z_18c = TensorMap(
    'seg_lax_4ch_64xy_50z_18c', Interpretation.CONTINUOUS, shape=(64, 64, 50, len(MRI_LAX_4CH_SEGMENTED_CHANNEL_MAP)+1),
    tensor_from_file=tff, channel_map=MRI_LAX_4CH_SEGMENTED_CHANNEL_MAP,
)

tff = _heart_mask_and_ecg_instances(
    'ukb_cardiac_mri', (64, 64, 50), 'cine_segmented_lax_4ch/2/',
    'cine_segmented_lax_4ch_annotated_', LAX_4CH_HEART_LABELS, ZeroMeanStd1(),
    'ukb_ecg_rest', (0, 0), ECG_REST_MEDIAN_LEADS, Standardize(mean=0, std=10),
    include_mri_segmentation=True,
)
seg_lax_4ch_128x_64y_50z_1c = TensorMap(
    'seg_lax_4ch_128x_64y_50z_1c', Interpretation.CONTINUOUS, shape=(128, 64, 50),
    tensor_from_file=tff, channel_map=MRI_LAX_4CH_SEGMENTED_CHANNEL_MAP,
)




# Imports
import os
import time
import logging
import numpy as np
from enum import Enum, auto
from itertools import chain
from abc import ABC, abstractmethod
from collections import defaultdict, Counter
from typing import Dict, List, Tuple, Iterable, Union, Optional, Set, Sequence, Callable, DefaultDict, Any

import tensorflow as tf
#import tensorflow_addons as tfa
from tensorflow.keras.layers import LeakyReLU, PReLU, ELU, ThresholdedReLU, Lambda, Reshape, LayerNormalization
from tensorflow.keras.callbacks import ModelCheckpoint, EarlyStopping, ReduceLROnPlateau, Callback
from tensorflow.keras.layers import SpatialDropout1D, SpatialDropout2D, SpatialDropout3D, add, concatenate
from tensorflow.keras.layers import Input, Dense, Dropout, BatchNormalization, Activation, Flatten, LSTM, RepeatVector
from tensorflow.keras.layers import Conv1D, Conv2D, Conv3D, UpSampling1D, UpSampling2D, UpSampling3D, MaxPooling1D
from tensorflow.keras.layers import MaxPooling2D, MaxPooling3D, Average, AveragePooling1D, AveragePooling2D, AveragePooling3D, Layer
from tensorflow.keras.layers import SeparableConv1D, SeparableConv2D, DepthwiseConv2D, Concatenate, Add
from tensorflow.keras.layers import GlobalAveragePooling1D, GlobalAveragePooling2D, GlobalAveragePooling3D
from tensorflow.keras.regularizers import L1, L2


Tensor = tf.Tensor

ACTIVATION_CLASSES = {
    'leaky': LeakyReLU(),
    'prelu': PReLU(),
    'elu': ELU(),
    'thresh_relu': ThresholdedReLU,
}
ACTIVATION_FUNCTIONS = {
    'swish': tf.keras.activations.swish,
    'gelu': tf.keras.activations.gelu,
    'mish': tf.keras.activations.mish,
}
NORMALIZATION_CLASSES = {
    'batch_norm': BatchNormalization,
    'layer_norm': LayerNormalization,
   # 'instance_norm': tfa.layers.InstanceNormalization,
   # 'poincare_norm': tfa.layers.PoincareNormalize,
}

CONV_REGULARIZATION_CLASSES = {
    # class name -> (dimension -> class)
    'spatial_dropout': {2: SpatialDropout1D, 3: SpatialDropout2D, 4: SpatialDropout3D},
    'dropout': defaultdict(lambda _: Dropout),
    'l1': L1,
    'l2': L2,
}
DENSE_REGULARIZATION_CLASSES = {
    'dropout': Dropout,
    'dropout': Dropout,
    'l1': L1,
    'l2': L2,
}


def _one_by_n_kernel(dimension):
    return tuple([1] * (dimension - 1))


def _conv_layer_from_kind_and_dimension(
        dimension: int, conv_layer_type: str, conv_x: List[int], conv_y: List[int], conv_z: List[int],
) -> Tuple[Layer, List[Tuple[int, ...]]]:
    if dimension == 4 and conv_layer_type == 'conv':
        conv_layer = Conv3D
        kernel = zip(conv_x, conv_y, conv_z)
    elif dimension == 3 and conv_layer_type == 'conv':
        conv_layer = Conv2D
        kernel = zip(conv_x, conv_y)
    elif dimension == 2 and conv_layer_type == 'conv':
        conv_layer = Conv1D
        kernel = zip(conv_x)
    elif dimension == 3 and conv_layer_type == 'separable':
        conv_layer = SeparableConv2D
        kernel = zip(conv_x, conv_y)
    elif dimension == 2 and conv_layer_type == 'separable':
        conv_layer = SeparableConv1D
        kernel = zip(conv_x)
    elif dimension == 3 and conv_layer_type == 'depth':
        conv_layer = DepthwiseConv2D
        kernel = zip(conv_x, conv_y)
    else:
        raise ValueError(f'Unknown convolution type: {conv_layer_type} for dimension: {dimension}')
    return conv_layer, list(kernel)


def _pool_layers_from_kind_and_dimension(dimension, pool_type, pool_number, pool_x, pool_y, pool_z):
    if dimension == 4 and pool_type == 'max':
        return [MaxPooling3D(pool_size=(pool_x, pool_y, pool_z)) for _ in range(pool_number)]
    elif dimension == 3 and pool_type == 'max':
        return [MaxPooling2D(pool_size=(pool_x, pool_y)) for _ in range(pool_number)]
    elif dimension == 2 and pool_type == 'max':
        return [MaxPooling1D(pool_size=pool_x) for _ in range(pool_number)]
    elif dimension == 4 and pool_type == 'average':
        return [AveragePooling3D(pool_size=(pool_x, pool_y, pool_z)) for _ in range(pool_number)]
    elif dimension == 3 and pool_type == 'average':
        return [AveragePooling2D(pool_size=(pool_x, pool_y)) for _ in range(pool_number)]
    elif dimension == 2 and pool_type == 'average':
        return [AveragePooling1D(pool_size=pool_x) for _ in range(pool_number)]
    else:
        raise ValueError(f'Unknown pooling type: {pool_type} for dimension: {dimension}')


def _upsampler(dimension, pool_x, pool_y, pool_z):
    if dimension == 4:
        return UpSampling3D(size=(pool_x, pool_y, pool_z))
    elif dimension == 3:
        return UpSampling2D(size=(pool_x, pool_y))
    elif dimension == 2:
        return UpSampling1D(size=pool_x)


def _activation_layer(activation: str) -> Activation:
    return (
        ACTIVATION_CLASSES.get(activation, None)
        or Activation(ACTIVATION_FUNCTIONS.get(activation, None) or activation)
    )


def _normalization_layer(norm: str) -> Layer:
    if not norm:
        return lambda x: x
    return NORMALIZATION_CLASSES[norm]()


def _regularization_layer(dimension: int, regularization_type: str, rate: float):
    if not regularization_type:
        return lambda x: x
    if regularization_type in DENSE_REGULARIZATION_CLASSES:
        return DENSE_REGULARIZATION_CLASSES[regularization_type](rate)
    return CONV_REGULARIZATION_CLASSES[regularization_type][dimension](rate)


def adaptive_normalize_from_tensor(tensor: Tensor, target: Tensor) -> Tensor:
    """Uses Dense layers to convert `tensor` to a mean and standard deviation to normalize `target`"""
    return adaptive_normalization(mu=Dense(target.shape[-1])(tensor), sigma=Dense(target.shape[-1])(tensor), target=target)


def adaptive_normalization(mu: Tensor, sigma: Tensor, target: Tensor) -> Tensor:
    target = tfa.layers.InstanceNormalization()(target)
    normalizer_shape = (1,) * (len(target.shape) - 2) + (target.shape[-1],)
    mu = Reshape(normalizer_shape)(mu)
    sigma = Reshape(normalizer_shape)(sigma)
    target *= sigma
    target += mu
    return target


def global_average_pool(x: Tensor) -> Tensor:
    if len(x.shape) == 3:
        return GlobalAveragePooling1D()(x)
    elif len(x.shape) == 4:
        return GlobalAveragePooling2D()(x)
    elif len(x.shape) == 5:
        return GlobalAveragePooling3D()(x)



from typing import Dict

import h5py
import numpy as np
from ml4h.TensorMap import TensorMap, Interpretation

ecg_5000_std = TensorMap('ecg_5000_std', Interpretation.CONTINUOUS, shape=(5000, 12))
ecg_single_lead_I = TensorMap(f'ecg_strip_I', Interpretation.CONTINUOUS, shape=(5000, 1))

hypertension_icd_only = TensorMap(
    name='hypertension_icd_only', interpretation=Interpretation.CATEGORICAL,
    channel_map={'no_hypertension_icd_only': 0, 'hypertension_icd_only': 1},
)
hypertension_icd_bp = TensorMap(
    name='hypertension_icd_bp', interpretation=Interpretation.CATEGORICAL,
    channel_map={'no_hypertension_icd_bp': 0, 'hypertension_icd_bp': 1},
)
hypertension_icd_bp_med = TensorMap(
    name='hypertension_icd_bp_med', interpretation=Interpretation.CATEGORICAL,
    channel_map={'no_hypertension_icd_bp_med': 0, 'hypertension_icd_bp_med': 1},
)
hypertension_med = TensorMap(
    name='start_fu_hypertension_med', interpretation=Interpretation.CATEGORICAL,
    channel_map={'no_hypertension_medication': 0, 'hypertension_medication': 1},
)

lvef = TensorMap(name='LVEF', interpretation=Interpretation.CONTINUOUS, channel_map={'LVEF': 0})

age = TensorMap(name='age_in_days', interpretation=Interpretation.CONTINUOUS, channel_map={'age_in_days': 0})
sex = TensorMap(name='sex', interpretation=Interpretation.CATEGORICAL, channel_map={'Female': 0, 'Male': 1})

cad = TensorMap(name='cad', interpretation=Interpretation.CATEGORICAL, channel_map={'no_cad': 0, 'cad': 1})
dm = TensorMap(name='dm', interpretation=Interpretation.CATEGORICAL, channel_map={'no_dm': 0, 'dm': 1})
hypercholesterolemia = TensorMap(
    name='hypercholesterolemia', interpretation=Interpretation.CATEGORICAL,
    channel_map={'no_hypercholesterolemia': 0, 'hypercholesterolemia': 1},
)

n_intervals = 25
af_tmap = TensorMap('survival_curve_af', Interpretation.SURVIVAL_CURVE, shape=(n_intervals*2,))
death_tmap = TensorMap('death_event', Interpretation.SURVIVAL_CURVE, shape=(n_intervals*2,))


def ecg_median_biosppy(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    tensor = np.zeros(tm.shape, dtype=np.float32)
    for lead in tm.channel_map:
        tensor[:, tm.channel_map[lead]] = hd5[f'{tm.path_prefix}{lead}']
    tensor = np.nan_to_num(tensor)
    return tensor

ecg_channel_map = {
    'I': 0, 'II': 1, 'III': 2, 'aVR': 3, 'aVL': 4, 'aVF': 5,
    'V1': 6, 'V2': 7, 'V3': 8, 'V4': 9, 'V5': 10, 'V6': 11,
}

ecg_biosppy_median_60bpm = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='median_60bpm_', shape=(600, 12),
    tensor_from_file=ecg_median_biosppy,
    channel_map=ecg_channel_map,
)
